/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */


/************************************************************************
 * thread.c: interface to thread library
************************************************************************/

#include "threadP.h"



/************************************************************************
 * global variables.
************************************************************************/

static char     printf_buf[PRINTF_BUFSIZE];






/************************************************************************
 * prototypes for local functions.
************************************************************************/

static	void	save_status(thread_t waitee, thread_t waiter);



/************************************************************************
 * interface functions
************************************************************************/

/************************************************************************
 * int thr_create(Tcb::Func func, void* args, unsigned flags,
 *     void* stack, unsigned stack_size, thread_t* tidp)
 *
 * create a new thread, initialize its state and make it runnable.
 * if `stack' and `stack_size' are specified, these parameters are used
 * as a stack for the thread, else a default one is allocated.
 *
 * `func' and `args' are the thread start function and its argument.
 * to the thread it appears as if `func(args)' had been invoked in the
 * new context.
 *
 * If 'flags & THR_', creates the thread but suspends it initially.
 * The thread is available to be scheduled when some other thread
 * calls 'thr_continue' for the suspended thread.
 *
 * if `tidp' is non-null, the id of the new thread is returned to the
 * caller thru `*tidp'.
************************************************************************/

int
thr_create(
    void*     stack,
    unsigned  stack_size,
	Tcb::Func func,
    void*     args,
    unsigned  flags,
    thread_t* tidp)
{
    Thread   t;


    THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }
    assert(thr_curr);

	/* construct a new thread */
	t = Tcb::Alloc( thr_curr->id.tid,
			func,
			args,
			flags,
			stack,
			stack_size,
			0 );

    (void) THR_SAVE_CONTEXT(t->exec.state);
    thr_do_arch_magic(t);

    if (tidp) {
        *tidp = t->id.tid;
    }

    DO_TRACE_0(("thr_create(): new t%u created, tcb %p", t->id.tid, t));

    if (flags & THR_SUSPENDED) {
        if (!tidp) {
            (void) thr_report_error("*** new thread may become garbage ***\n");
        }
    }
    else {
        ENQUEUE(thr_ready_q, t);
    }

    CHECK_RESCHEDULE();

    THR_LEAVE_LIB();
    return thr_err(THR_OKAY);
}





/************************************************************************
 * void thr_yield(void)
 *
 * give up control to the scheduler. the calling thread is scheduled
 * to run as soon as the currently ready threads have been scheduled.
************************************************************************/

void
thr_yield(void)
{
    if (!thr_init_done) {
		thr_report_error("*** thread library not initialized ***\n");
        return;
    }

    THR_ENTER_LIB();

    assert(thr_curr);

    DO_TRACE_0(("thr_yield():"));

    ENQUEUE(thr_ready_q, thr_curr);
    thr_reschedule();

    THR_LEAVE_LIB();
    return;
}





/************************************************************************
 * void thr_cleanup(void* result)
 * void thr_exit(void* result)
 *
 * terminate the calling thread and save `result' as the return value.
 * checks for any threads that are waiting for the current thread
 * or any thread. the waiter is scheduled to run.
 *
 * if no waiter is found, the current thread is zombied.
 *
 * all files bound to the dying thread are released and any queued
 * messages are discarded.
 *
 * `thr_cleanup()' is the internal function that does the actual
 * work of bringing down the exiting thread. `thr_exit()' is simply
 * a wrapper around this function. functions within libthread should
 * not call the internal function.
************************************************************************/

void
thr_cleanup(void* result)
{
    thread_t waiter;
    Thread   t;
    unsigned i;


    THR_ENTER_LIB();

    assert(thr_init_done);
    assert(thr_curr);

    thr_curr->flags.exited = 1;
    thr_curr->call.result  = result;

    if (thr_curr->flags.waiting) {
        waiter = thr_curr->wait.waiter;
        save_status(thr_curr->id.tid, waiter);

        DO_TRACE_1(("thr_cleanup(): found registered waiter, enabling t%u",
            waiter));
        ENQUEUE(thr_ready_q, thr_tab->GetThread(waiter));
    }
    else {
        if (ISEMPTYQ(thr_waiter_q)) {
            DO_TRACE_1(("thr_cleanup(): no waiters, becoming a zombie"));
            ENQUEUE(thr_zombie_q, thr_curr);
        }
        else {
            t = (Thread) DEQUEUE(thr_waiter_q);
            save_status(thr_curr->id.tid, t->id.tid);
            DO_TRACE_1(("thr_cleanup(): found anonymous waiter, enabling t%u",
                t->id.tid));
            ENQUEUE(thr_ready_q, t);
        }
    }

    for (i = 0; i < THR_KEY_TABLE_SIZE; i++) {
        if ((thr_key_destructors[i] != 0)
            && (thr_curr->key.key_data[i] != 0)) {
            (thr_key_destructors[i])(thr_curr->key.key_data[i]);
            thr_curr->key.key_data[i] = 0;
        }
    }

	thr_tab->UnbindAll();

    if (thr_curr->msg.mq) {
        FREEQ(thr_curr->msg.mq);
    }

    THR_LEAVE_LIB();
    return;
}



void
thr_exit(void* result)
{
    thr_cleanup(result);
    thr_reschedule();

    /* NOTREACHED */
	thr_report_error("*** thread returned after exit ***\n");
    abort();
    return;
}





/************************************************************************
 * thread_t thr_self(void)
 *
 * return the id of the calling thread.
************************************************************************/

thread_t
thr_self(void)
{
    THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }
    assert(thr_curr);

    CHECK_REALLY_RESCHEDULE();

    THR_LEAVE_LIB();
    return thr_curr->id.tid;
}





/************************************************************************
 * thread_t thr_parent(void)
 *
 * return the id of the parent of the calling thread.
************************************************************************/

thread_t
thr_parent(void)
{
    THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }
    assert(thr_curr);

    CHECK_REALLY_RESCHEDULE();

    THR_LEAVE_LIB();
    return thr_curr->id.ptid;
}





/************************************************************************
 * const char* thr_name(const char* new_name)
 *
 * set the string name of the current thread to `new_name' and return
 * the current name. if `new_name' is null, the old name is unchanged.
 * names are truncated to THR_NAMELEN characters.
************************************************************************/

const char*
thr_name(const char* new_name)
{
    THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }
    assert(thr_curr);

    if (new_name) {
        (void) strncpy(thr_curr->id.name, new_name, THR_NAMELEN);
        thr_curr->id.name[THR_NAMELEN] = '\0';
    }

    CHECK_REALLY_RESCHEDULE();

    THR_LEAVE_LIB();
    return thr_curr->id.name;
}





/************************************************************************
 * int thr_join(thread_t waitfor, thread_t* departed, void** status)
 *
 * wait for the termination of thread `waitfor' or if 0, any thread.
 * the id of the actual thread reaped is returned via `*departed'
 * its return value via `*status'.
************************************************************************/

int
thr_join(thread_t waitfor, thread_t* departed, void** status)
{
    int ret;

    if (!thr_init_done) {
		thr_report_error("*** thread library not initialized ***\n");
        return thr_err(THR_ERR);
    }

    THR_ENTER_LIB();

    assert(thr_curr);

    DO_TRACE_0(("thr_join(%u):", waitfor));

    if ((waitfor != 0) && (waitfor == thr_curr->id.tid)) {
        thr_report_error("cannot wait for self---deadlock?\n");
        RETURN(THR_EDEADLK);
    }

    if ((waitfor != 0) && !thr_tab->IsThread(waitfor)) {
        thr_report_error("invalid tid `%u' specified as waitfor\n", waitfor);
        RETURN(THR_ESRCH);
    }

    thr_curr->wait.departed = departed;
    thr_curr->wait.status   = status;

    if (waitfor != 0) {
        Thread t;

        t = thr_tab->GetThread(waitfor);

        if (t->flags.reaped) {
            thr_report_error("t%u already reaped\n", waitfor);
            RETURN(THR_ESRCH);
        }
        if (t->flags.waiting) {
            thr_report_error("t%u is waited for by t%u\n",
                waitfor, t->wait.waiter);
            RETURN(THR_ESRCH);
        }

        if (t->flags.exited) {
            save_status(waitfor, thr_curr->id.tid);
            DO_TRACE_0(("thr_join(%u): waitee already exited, reaping now",
                waitfor));
        }
        else {
            DO_TRACE_0(("thr_join(%u): waitee still running, registering wait",
                waitfor));
            t->flags.waiting = 1;
            t->wait.waiter   = thr_curr->id.tid;

            thr_reschedule();
            RETURN(THR_OKAY);
        }
    }
    else {
        Thread   t;
        unsigned found;

        found = 0;
        while (!ISEMPTYQ(thr_zombie_q)) {
            t = (Thread) DEQUEUE(thr_zombie_q);
            if (!(t->flags.reaped)) {
                DO_TRACE_0(("thr_join(%u): found reapable zombie t%u",
                    waitfor, t->id.tid));
                save_status(t->id.tid, thr_curr->id.tid);
                found = 1;
                break;
            }
        }

        if (!found) {
            DO_TRACE_0(("thr_join(%u): no reapable zombie, going into waiter queue",
                waitfor));
            ENQUEUE(thr_waiter_q, thr_curr);
            thr_reschedule();
            RETURN(THR_OKAY);
        }
    }

    CHECK_RESCHEDULE();
    ret = thr_err(THR_OKAY);

finish:
    THR_LEAVE_LIB();
    return ret;
}





/************************************************************************
 * int thr_kill(thread_t target, int sig)
 *
 * send signal `sig' to thread `target'. currently all signals kill
 * the target thread.
************************************************************************/

int
thr_kill(thread_t target, int sig)
{
    int    ret;
    Thread t;

    if (!thr_init_done) {
        thr_report_error("*** thread library not initialized ***\n");
        return THR_ERR;
    }

    THR_ENTER_LIB();

    assert(thr_curr);

    if (!thr_tab->IsThread(target)) {
        thr_report_error("invalid tid `%u' as target\n", target);
        RETURN(THR_ESRCH);
    }

    t = thr_tab->GetThread(target);

    if (t->flags.exited) {
        thr_report_error("cannot kill dead t%u\n", target);
        RETURN(THR_ESRCH);
    }

    DO_TRACE_0(("thr_kill(%u,%d):", target, sig));
    t->flags.killed = 1;

    CHECK_REALLY_RESCHEDULE();
    ret = thr_err(THR_OKAY);

finish:
    THR_LEAVE_LIB();
    return ret;
}





/************************************************************************
 * int thr_suspend(thread_t target)
 *
 * suspend thread `target'. the thread will be rescheduled only after
 * it is explicitly continued.
 *
 *
 * int thr_continue(thread_t target)
 *
 * continues a suspended thread `target'.
************************************************************************/

int
thr_suspend(thread_t target)
{
    int    ret;
    Thread t;

    if (!thr_init_done) {
        thr_report_error("*** thread library not initialized ***\n");
        return THR_ERR;
    }

    THR_ENTER_LIB();

    assert(thr_curr);

    if (!thr_tab->IsThread(target)) {
        thr_report_error("invalid tid `%u' as target\n", target);
        RETURN(THR_ESRCH);
    }

    t = thr_tab->GetThread(target);

    if (t->flags.exited) {
        thr_report_error("cannot suspend dead t%u\n", target);
        RETURN(THR_ESRCH);
    }

    DO_TRACE_0(("thr_suspend(%u):", target));
    t->flags.suspended = 1;

    CHECK_REALLY_RESCHEDULE();
    ret = thr_err(THR_OKAY);

finish:
    THR_LEAVE_LIB();
    return ret;
}



int
thr_continue(thread_t target)
{
    int    ret;
    Thread t;

    if (!thr_init_done) {
        thr_report_error("*** thread library not initialized ***\n");
        return THR_ERR;
    }

    THR_ENTER_LIB();

    assert(thr_curr);

    if (!thr_tab->IsThread(target)) {
        thr_report_error("invalid tid `%u' as target\n", target);
        RETURN(THR_ESRCH);
    }

    t = thr_tab->GetThread(target);

    if (t->flags.exited) {
        thr_report_error("cannot continue dead t%u\n", target);
        RETURN(THR_ESRCH);
    }

    if (!t->flags.suspended) {
        thr_report_error("cannot continue non-suspended t%u\n", target);
        RETURN(THR_ESRCH);
    }

    DO_TRACE_0(("thr_continue(%u):", target));
    t->flags.suspended = 0;
    ENQUEUE(thr_ready_q, t);

    CHECK_REALLY_RESCHEDULE();
    ret = thr_err(THR_OKAY);

finish:
    THR_LEAVE_LIB();
    return ret;
}





/************************************************************************
 * int thr_keycreate(thread_key_t* keyp, void (*destructor)(void *)
 *
 * create a new key in the thread library and install it.
************************************************************************/

int
thr_keycreate(thread_key_t* keyp, void (*destructor)(void *))
{
    int ret;

    THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }

    if (keyp == 0) {
        thr_report_error("null key pointers not allowed");
        RETURN(THR_ERR);
    }

    if (thr_curr_key == THR_KEY_TABLE_SIZE) {
        thr_report_error("out of key space, recompile with larger size\n");
        RETURN(THR_ENOMEM);
    }

    *keyp = thr_curr_key++;
    thr_key_destructors[*keyp] = (Destructor) destructor;

    CHECK_REALLY_RESCHEDULE();
    ret = thr_err(THR_OKAY);

finish:
    THR_LEAVE_LIB();
    return ret;
}





/************************************************************************
 * int thr_setspecific(thread_key_t key, void* data)
 *
 * load `data' into the local storage of the current thread with the
 * given key of `key'.
************************************************************************/

int
thr_setspecific(thread_key_t key, void* data)
{
    int ret;

    if (!thr_init_done) {
        thr_report_error("*** thread library not initialized ***\n");
        return THR_ERR;
    }

    THR_ENTER_LIB();

    assert(thr_curr);

    if (is_invalid_key(key)) {
        RETURN(THR_EBADKEY);
    }

    thr_curr->key.key_data[key] = data;

    CHECK_REALLY_RESCHEDULE();
    ret = thr_err(THR_OKAY);

finish:
    THR_LEAVE_LIB();
    return ret;
}





/************************************************************************
 * int thr_getspecific(thread_key_t key, void** datap)
 *
 * set `*datap' to the value of a previously stored data object with the
 * given key of `key'.
************************************************************************/

int
thr_getspecific(thread_key_t key, void** datap)
{
    int ret;

    if (!thr_init_done) {
        thr_report_error("*** thread library not initialized ***\n");
        return THR_ERR;
    }

    THR_ENTER_LIB();

    assert(thr_curr);

    if (is_invalid_key(key)) {
        RETURN(THR_EBADKEY);
    }

    if (datap == 0) {
        RETURN(THR_ERR);
    }

    *datap = thr_curr->key.key_data[key];

    CHECK_REALLY_RESCHEDULE();
    ret = thr_err(THR_OKAY);

finish:
    THR_LEAVE_LIB();
    return ret;
}





/************************************************************************
 * void thr_perror(const char* msg)
 *
 * libthread's version of perror. the supported error codes are those
 * exported via thread.h.
************************************************************************/

void
thr_perror(const char* msg)
{
    if (!thr_init_done) {
        thr_report_error("*** thread library not initialized ***\n");
        return;
    }

    THR_ENTER_LIB();

    assert(thr_curr);
    assert(thr_curr->id.errno_local <= 0);
    assert(-(thr_curr->id.errno_local) < (int) thr_nerr);

    (void) thr_report_error("%s: %s\n", msg,
        thr_errlst[-(thr_curr->id.errno_local)]);

    THR_LEAVE_LIB();
    return;
}





/************************************************************************
 * void thr_trace_on(void)
 * void thr_trace_off(void)
 *
 * programmatic control for thread library debugging and tracing.
 * the thread library clients can use these functions to enable and
 * disable tracing at appropriate points in the code. this method is
 * orthogonal to environment variables.
************************************************************************/

void
thr_trace_on(void)
{
#if defined(ENABLE_TRACE)
    thr_trace_flag = 2;
#else /* defined(ENABLE_TRACE) */
    (void) thr_report_error("trace support not compiled in\n");
#endif /* defined(ENABLE_TRACE) */

    return;
}



void
thr_trace_off(void)
{
#if defined(ENABLE_TRACE)
    thr_trace_flag = 0;
#else /* defined(ENABLE_TRACE) */
    (void) thr_report_error("trace support not compiled in\n");
#endif /* defined(ENABLE_TRACE) */

    return;
}





/************************************************************************
 * void thr_do_trace(const char* format, ...)
 * void thr_do_trace_0(const char* format, ...)
 * void thr_do_trace_1(const char* format, ...)
 *
 * printf-like trace functions for the thread library. depending on
 * the value of the global variable `thr_trace_flag' different functions
 * are triggered.
 *
 * the traces are always prefixed with the <pid,tid,tname> triple for the
 * thread in question. traces are appended to a trace file that is
 * set during the initialization of the thread library---it is stderr
 * by default.
************************************************************************/

void
thr_do_trace_aux(const char* format, va_list& args )
{
    (void) fflush(stdout);
    (void) fflush(stderr);

    (void) vsprintf(printf_buf, format, args);
    (void) fprintf(thr_trace_fp, "[%u,%u,%s] %s\n",
        thr_pid, thr_curr->id.tid, thr_curr->id.name, printf_buf);
    (void) fflush(thr_trace_fp);
}


void
thr_do_trace(const char* format, ...)
{
	va_list args;


	THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }

	va_start( args, format );
	thr_do_trace_aux( format, args );
	va_end( args );

    THR_LEAVE_LIB();
    return;
}

void
thr_do_trace_0(const char* format, ...)
{
	va_list args;

    if (thr_trace_flag < 1) {
        return;
    }

	THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }

	va_start( args, format );
	thr_do_trace_aux( format, args );
	va_end( args );

	THR_LEAVE_LIB();
    return;
}


void
thr_do_trace_1(const char* format, ...)
{
	va_list args;

    if (thr_trace_flag < 2) {
        return;
    }

	THR_ENTER_LIB();

    if (!thr_init_done) {
        thr_init();
    }

	va_start( args, format );
	thr_do_trace_aux( format, args );
	va_end( args );

	THR_LEAVE_LIB();
    return;
}





/************************************************************************
 * type and value discovery functions
************************************************************************/


/************************************************************************
 * item_t thr_type(thread_t tid)
 *
 * Returns the type of the item represented by the given 'tid.'  Returns
 * item_t_unknown if tid is not valid.
************************************************************************/

item_t
thr_type(thread_t tid)
{
	item_t ret = item_t_unknown;

	if( thr_tab->IsValid(tid) )
	{
		ret = thr_tab->GetType(tid);
	}

	return ret;
}


/************************************************************************
 * item_t thr_type(thread_t tid)
 *
 * Returns the file descriptor associated with the item named by 'tid.'
************************************************************************/

PDDESC
thr_file(thread_t tid)
{
	PDDESC ret = INVALID_PDDESC;

	if( thr_tab->IsValid(tid) && thr_tab->IsFile(tid) )
	{
		ret = thr_tab->GetFile(tid)->fd;
	}

	return ret;
}


/************************************************************************
 * item_t thr_type(thread_t tid)
 *
 * Returns the socket descriptor associated with the item named by 'tid.'
************************************************************************/

PDSOCKET
thr_socket(thread_t tid)
{
	PDSOCKET ret = INVALID_PDSOCKET;

	if( thr_tab->IsValid(tid) && thr_tab->IsSocket(tid) )
	{
		ret = thr_tab->GetSocket(tid)->sock;
	}

	return ret;
}



/************************************************************************
 * utility functions
************************************************************************/





/************************************************************************
 * void save_status(thread_t waitee, thread_t waiter)
 *
 * `waitee' has finished execution and `waiter' is joining the `waitee'.
 * save the id of the exiting thread and its result in the waiter's
 * thread control block.
************************************************************************/

static
void
save_status(thread_t waitee, thread_t waiter)
{
    Thread t_waitee;
    Thread t_waiter;


    assert(thr_tab->IsThread(waitee));
    assert(thr_tab->IsThread(waiter));

    t_waitee = thr_tab->GetThread(waitee);
    t_waiter = thr_tab->GetThread(waiter);

    t_waitee->flags.reaped = 1;
    if (t_waiter->wait.departed) {
        *(t_waiter->wait.departed) = waitee;
    }

    if (t_waiter->wait.status) {
        *(t_waiter->wait.status) = t_waitee->call.result;
    }

    return;
}


